
/* Builds call stats from the openser log  */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <limits.h>
#include <errno.h>

#undef DBG
/*#define DBG 1*/

#define CMDDATEFMT "%Y %b %d %H:%M:%S"
#define LOGDATEFMT "%b %d %H:%M:%S"
#define DEFAULTCALLSPERUNIT 1000000LLU

#define ANSWER_DEADLINE 1000000ULL /* in microseconds */

#define LINE_BUFFER_SIZE 1024
#define USERNAME_SIZE 128
#define URI_SIZE 128
#define CID_SIZE 128
#define OPENSER_PNAME "openser"

typedef uint64_t log_timestamp_t;
typedef uint64_t log_timeinterval_t;

/* note: lookups are ordered by their logged timestamps */
struct lookup {
	struct lookup *next_lookup;
	log_timestamp_t timestamp; /* event timestamp */
	char username[USERNAME_SIZE]; /* looked up who */
	log_timeinterval_t time; /* time the lookup took */
};

struct lookup *lookups;
struct lookup *prev_added_lookup_pred;
/* accounted call events */
#define INVITE_RECEIVED	0
#define INVITE_REPLIED	1
#define ACK_RECEIVED	2 /* received?!? not sure*/
#define BYE_REPLIED	3
#define NUM_EVENTS	(BYE_REPLIED+1)
#define ALL_EVENTS_MASK	((1 << NUM_EVENTS)-1)
struct call_event {
	log_timestamp_t timestamp;
	int count; /* how many of this have we received (repeats) */
	int code;
};
struct call {
	struct call *next_call;
	char caller_uri[URI_SIZE];
	char receiver_uri[URI_SIZE];
	char call_id[CID_SIZE];
	unsigned int call_events;
	struct call_event event[NUM_EVENTS]; 
};
struct call *calls;
struct call *completed_calls;

/* TODO: prol'ly no point in saving these */
struct log_event {
/*	time_t	event_time;*/
	char	month[4];
	char	day[3];
	char	time[9];
	char	host[64];
	char	process[64];
	char	message[1024];
};

static unsigned long long num_lines;
static unsigned long long num_lookups;
static unsigned long long num_calls;
static unsigned long long num_completed_calls;

static unsigned int cmd_option_quiet = 0;
static unsigned int cmd_option_list_calls = 0;
static struct tm *cmd_option_begin_date = NULL;
static struct tm *cmd_option_end_date = NULL;
static char cmd_option_output_basename[512];
static FILE *cmd_option_call_output = NULL;
static FILE *cmd_option_cps_output = NULL;
static FILE *cmd_option_lookup_output = NULL;
static unsigned int cmd_option_output_csv = 0;
static unsigned int cmd_option_drop_tsfaults = 0;
static log_timestamp_t cmd_option_calls_per_unit = 0ULL;

static void print_call(FILE *fd, struct call *cp) {
	int i;
	fprintf(fd, "%s %s %s (ev: %04x)\n", cp->call_id, cp->caller_uri, cp->receiver_uri, cp->call_events);
	for (i = 0; i < NUM_EVENTS; i++) {
		if (cp->call_events & (1 << i))
			fprintf(fd, "      (%04x) %llu n: %d code: %d\n",
				(1 << i), 
				cp->event[i].timestamp, 
				cp->event[i].count, cp->event[i].code);
	}
}

/*
 * stats generation
 */
static void generate_lookup_csv_files() {
	struct lookup *lp;
	
	/* output calls */
	lp = lookups;
	while (lp != NULL) {
		if (cmd_option_lookup_output) {
			if (cmd_option_output_csv)
				fprintf(cmd_option_lookup_output, 
					"%llu,\"%s\",%llu\n",
					lp->timestamp, lp->username,
					lp->time);
			else
				fprintf(cmd_option_lookup_output, 
					"%llu %llu\n",
					lp->timestamp,
					lp->time);
		}
		lp = lp->next_lookup;
	}
}

static void generate_lookup_stats() {
	struct lookup *lp;
	double avg_lookup_time;
	long long lookup_count;
	log_timestamp_t first_lookup_ts=0, last_lookup_ts=0;

	lookup_count = 0;
	avg_lookup_time = 0.0;
	lp = lookups;
	if (lp != NULL)
		first_lookup_ts = lp->timestamp;
#ifdef DBG
	printf("\nlookups @%p:\n\n", lookups);
#endif
	while (lp != NULL) {
		avg_lookup_time += (double)lp->time;
		lookup_count++;	
		last_lookup_ts = lp->timestamp;
		lp = lp->next_lookup;
	}
	/* csv file output */
	if (cmd_option_lookup_output) {
		generate_lookup_csv_files();
	}
	if (lookup_count > 0) {
		avg_lookup_time = avg_lookup_time/(double)lookup_count;
		printf("  lookups done in period [%llu - %llu] = %llu us\n", 
			first_lookup_ts, last_lookup_ts, 
			last_lookup_ts-first_lookup_ts);
		printf("  average lookup time: %f us\n", avg_lookup_time);
	}
}

static int check_call_timestamp_order(struct call *cp) {
	if (cp->event[INVITE_RECEIVED].timestamp > 
		cp->event[INVITE_REPLIED].timestamp)
		return 1;	
	if (cp->event[INVITE_REPLIED].timestamp > 
		cp->event[ACK_RECEIVED].timestamp)
		return 1;	
	if (cp->event[ACK_RECEIVED].timestamp > 
		cp->event[BYE_REPLIED].timestamp)
		return 1;	
	return 0;
}

static void generate_call_csv_files() {
	long long failed_per_sec, deadlined_per_sec, completed_per_sec;
	log_timestamp_t current_time, first_time, last_time, time_step;
	struct call *cp;
	
	/* output calls */
	cp = completed_calls;
	while (cp != NULL) {
		if (cmd_option_call_output) {
			if (cmd_option_drop_tsfaults && 
					check_call_timestamp_order(cp)) {
				cp = cp->next_call;
				continue;
			}
			if (cmd_option_output_csv)
				fprintf(cmd_option_call_output, 
				"%llu,\"%s\",%04x,%llu,%llu,%llu,%llu\n",
				cp->event[0].timestamp, cp->call_id,
				cp->call_events,
				cp->event[BYE_REPLIED].timestamp -
				cp->event[INVITE_RECEIVED].timestamp,
				cp->event[INVITE_REPLIED].timestamp -
				cp->event[INVITE_RECEIVED].timestamp,
				cp->event[ACK_RECEIVED].timestamp -
				cp->event[INVITE_REPLIED].timestamp,
				cp->event[BYE_REPLIED].timestamp -
				cp->event[ACK_RECEIVED].timestamp);
			else
				fprintf(cmd_option_call_output, 
				"%llu %04u %llu %llu %llu %llu\n",
				cp->event[0].timestamp, 
				cp->call_events,
				cp->event[BYE_REPLIED].timestamp -
				cp->event[INVITE_RECEIVED].timestamp,
				cp->event[INVITE_REPLIED].timestamp -
				cp->event[INVITE_RECEIVED].timestamp,
				cp->event[ACK_RECEIVED].timestamp -
				cp->event[INVITE_REPLIED].timestamp,
				cp->event[BYE_REPLIED].timestamp -
				cp->event[ACK_RECEIVED].timestamp);
		}
		cp = cp->next_call;
	}
	cp = calls;
	while (cp != NULL) {
		if (cmd_option_call_output) {
			log_timestamp_t t1, t2, t3, t4;
			if (cp->call_events & (1 << BYE_REPLIED) &&
				cp->call_events & (1 << INVITE_RECEIVED))
				t1 = cp->event[BYE_REPLIED].timestamp -
					cp->event[INVITE_RECEIVED].timestamp;
			else
				t1 = 0ULL;
			if (cp->call_events & (1 << INVITE_REPLIED) &&
				cp->call_events & (1 << INVITE_RECEIVED))
				t2 = cp->event[INVITE_REPLIED].timestamp -
					cp->event[INVITE_RECEIVED].timestamp;
			else
				t2 = 0ULL;
			if (cp->call_events & (1 << ACK_RECEIVED) &&
				cp->call_events & (1 << INVITE_REPLIED))
				t3 = cp->event[ACK_RECEIVED].timestamp -
					cp->event[INVITE_REPLIED].timestamp;
			else
				t3 = 0ULL;
			if (cp->call_events & (1 << BYE_REPLIED) &&
				cp->call_events & (1 << ACK_RECEIVED))
				t4 = cp->event[BYE_REPLIED].timestamp -
					cp->event[ACK_RECEIVED].timestamp;
			else
				t4 = 0ULL;

			if (cmd_option_output_csv)
				fprintf(cmd_option_call_output, 
				"%llu,\"%s\",%04x,%llu,%llu,%llu,%llu\n",
				cp->event[0].timestamp, cp->call_id,
				cp->call_events,
				t1, t2, t3, t4);
			else
				fprintf(cmd_option_call_output, 
				"%llu %04u %llu %llu %llu %llu\n",
				cp->event[0].timestamp, 
				cp->call_events,
				t1, t2, t3, t4);

			if ((cp->call_events & (1 << INVITE_RECEIVED)) == 0)
				fprintf(stderr, "warning: incomplete call without INVITE (%s)!\n", cp->call_id);
		}
		cp = cp->next_call;
	}

	/* output cps */
	/* find first and last starting event timestamp */
	current_time = UINT64_MAX;
	last_time = 0ULL;
	cp = completed_calls;
	while (cp != NULL) {
		if (cmd_option_drop_tsfaults && 
				check_call_timestamp_order(cp)) {
			cp = cp->next_call;
			continue;
		}
		if (cp->event[0].timestamp < current_time)
			current_time = cp->event[0].timestamp;
		if (cp->event[0].timestamp > last_time)
			last_time = cp->event[0].timestamp;
		cp = cp->next_call;
	}
	cp = calls;
	while (cp != NULL) {
		if (cp->call_events & (1 << INVITE_RECEIVED) && 
				cp->event[0].timestamp < current_time)
			current_time = cp->event[0].timestamp;
		if (cp->call_events & (1 << INVITE_RECEIVED) &&
				cp->event[0].timestamp > last_time)
			last_time = cp->event[0].timestamp;
		cp = cp->next_call;
	}
	first_time = current_time;
	if (cmd_option_calls_per_unit != 0)
		time_step = cmd_option_calls_per_unit;
	else
		time_step = DEFAULTCALLSPERUNIT;
	while (current_time <= last_time) {
		failed_per_sec = deadlined_per_sec = completed_per_sec = 0;
		cp = completed_calls;
		while (cp != NULL) {
			if (cmd_option_drop_tsfaults && 
					check_call_timestamp_order(cp)) {
				cp = cp->next_call;
				continue;
			}
			if (current_time <= cp->event[0].timestamp &&
					cp->event[0].timestamp < 
					current_time+time_step) {
				if (cp->event[INVITE_REPLIED].timestamp -
					cp->event[INVITE_RECEIVED].timestamp >
					ANSWER_DEADLINE)
					deadlined_per_sec++;
				else
					completed_per_sec++;
			}
			cp = cp->next_call;
		}	
		cp = calls;
		while (cp != NULL) {
			if (cp->call_events & (1 << INVITE_RECEIVED) &&
				 current_time <= cp->event[0].timestamp &&
					cp->event[0].timestamp < 
					current_time+time_step) {
				failed_per_sec++;
			}
			cp = cp->next_call;
		}
		if (cmd_option_output_csv)
			fprintf(cmd_option_cps_output, "%llu,%lld,%lld,%lld\n", 
				current_time-first_time, 
				completed_per_sec, 
				deadlined_per_sec, 
				failed_per_sec);	
		else
			fprintf(cmd_option_cps_output, "%llu %lld %lld %lld\n", 
				current_time-first_time, 
				completed_per_sec, 
				deadlined_per_sec, 
				failed_per_sec);	
		current_time += time_step;
	}
	
}

static void generate_call_stats() {
	struct call *cp;
	long long calls_completed, calls_incomplete, calls_deadlined, 
			calls_outoforderts;
	long long call_length;
	long long min_call_length, max_call_length;
	double avg_call_length;

	calls_completed = calls_incomplete = 
		calls_deadlined = calls_outoforderts = 0;
	min_call_length = INT64_MAX;
	max_call_length = 0;
	avg_call_length = 0.0;
	cp = completed_calls;
#ifdef DBG
	printf("\ncomplete_calls @%p:\n\n", cp);
#endif
	if (cmd_option_list_calls)
		printf("deadlined complete calls (details):\n");
	while (cp != NULL) {
		call_length = cp->event[BYE_REPLIED].timestamp -
			cp->event[INVITE_RECEIVED].timestamp;
		if (cp->event[INVITE_REPLIED].timestamp -
				cp->event[INVITE_RECEIVED].timestamp >
				ANSWER_DEADLINE) {
			calls_deadlined++;
			if (cmd_option_list_calls)
				print_call(stdout, cp);
		}
		if (check_call_timestamp_order(cp)) {
			calls_outoforderts++;
			fprintf(stderr, "warning: wrong order in call timestamps:\n");
			print_call(stderr, cp);
		}
		if (call_length < min_call_length)
			min_call_length = call_length;
		if (call_length > max_call_length)
			max_call_length = call_length;
		avg_call_length += call_length;
		calls_completed++;
		cp = cp->next_call;
	}
	cp = calls;
#ifdef DBG
	printf("\ncalls @%p:\n\n", cp);
#endif
	if (cmd_option_list_calls)
		printf("incomplete calls (details):\n");
	while (cp != NULL) {
		if (cmd_option_list_calls)
			print_call(stdout, cp);
		calls_incomplete++;
		cp = cp->next_call;
	}

	/* csv file output */
	if (cmd_option_call_output || cmd_option_cps_output) {
		generate_call_csv_files();
	}

	avg_call_length = avg_call_length / calls_completed;
	printf("  completed calls: %lld\n", calls_completed);
	printf("    calls with exceeded deadline (of %llu us): %lld\n", 
		ANSWER_DEADLINE, calls_deadlined);
	printf("    calls with out-of-order timestamps: %lld\n", 
		calls_outoforderts);
	printf("    min. call length: %lld us\n", min_call_length);
	printf("    max. call length: %lld us\n", max_call_length);
	printf("    avg. call length: %f us\n", avg_call_length);	
	printf("  incomplete calls: %lld\n", calls_incomplete);
}

/* remove repeated messages (retries) that belong to complete calls */
static void postprocess_calls() {
	struct call *cp, *prev_cp;
	struct call *ccp;
	unsigned long long n;
	unsigned long long cn;

	n = cn = 0;
	prev_cp = NULL;
	cp = calls;
	while (cp != NULL) {
		ccp = completed_calls;
		if (!cmd_option_quiet && (cn % 10) == 0) {
			printf("postprocessing calls [%llu]\r", cn);
			fflush(stdout);
		}
		while (ccp != NULL) {
			/* XXX receiver uri not checked... */
			if (!strcmp(cp->call_id, ccp->call_id) && 
				!strcmp(cp->caller_uri, ccp->caller_uri))
			{
				if (cp == calls) {
//					printf("pp: removed2 [%s]\n", cp->call_id);
					calls = cp->next_call;
					n++;	
					break;
				}
				else if (prev_cp) {
//					printf("pp: removed1 [%s]\n", cp->call_id);
					prev_cp->next_call = cp->next_call;
					n++;
					cp = prev_cp;
					break;
				}			
				else {
					fprintf(stderr, "warning: in call_postproc\n");
				}
			}
			ccp = ccp->next_call;
		}
		prev_cp = cp;
		cp = cp->next_call;
		cn++;
	}
#ifdef DBG
	printf("postprocess_calls: cleaned out %llu duplicates\n", n);
#endif
}

/*
 * lookup and call list handling
 */
static void add_lookup(log_timestamp_t ts, const char *username, 
		log_timeinterval_t t) {
	struct lookup *lp, *prev_lp, *ilp;
	struct lookup *newl;
	int i;	
/*	int tmpperf1 = 0, tmpperf2 = 0;*/

	/*printf("addlookup (%llu)\n", ts);*/
	if ((newl = (struct lookup *)malloc(sizeof(struct lookup))) == NULL) {
		fprintf(stderr, "add_lookup: Out of memory.\n");
		exit(1);
	}
	strcpy(newl->username, username);
	for (i = 0; i < strlen(username); i++)
		newl->username[i] = toupper(newl->username[i]);
	newl->timestamp = ts;
	newl->time = t;
	newl->next_lookup = NULL;

	/* find the proper spot by timestamp */
	/* try list tail first... */
	prev_lp = NULL;
	ilp = NULL;
	lp = prev_added_lookup_pred;
	while (lp != NULL) {
		/*printf("   ts1:%lld\n", lp->timestamp);*/
		if (lp->timestamp > newl->timestamp) {
			ilp = prev_lp;
			break;
		}
		prev_lp = lp;
		lp = lp->next_lookup;
		/*tmpperf1++;*/
	}
	/* ...and then the whole list */
	if (prev_added_lookup_pred == NULL || (ilp == NULL &&
			prev_lp == NULL && lp->timestamp > newl->timestamp)) {
		prev_lp = NULL;
		ilp = NULL;
		lp = lookups;
		while (lp != NULL) {
			/*printf("   ts2:%lld\n", lp->timestamp);*/
			if (lp->timestamp > newl->timestamp) {
				ilp = prev_lp;
				break;
			}
			prev_lp = lp;
			lp = lp->next_lookup;
			/*tmpperf2++;*/
		}
	}
	
	if (ilp == NULL) {
		if (prev_lp == NULL) { /* list head */
			/*printf("  head\n");*/
			if (lookups)
				newl->next_lookup = lookups->next_lookup;
			lookups = newl;
		}
		else { /* list tail */
			/*printf("  tail\n");*/
			prev_added_lookup_pred = prev_lp;
			prev_lp->next_lookup = newl;
		}
	}
	else {
		/*printf("  after: %llu\n", ilp->timestamp);*/
		prev_added_lookup_pred = ilp;
		newl->next_lookup = ilp->next_lookup;
		ilp->next_lookup = newl;
	}
	num_lookups++;
	/*printf("tmpperf1: %d tmpperf2: %d ts:%lld\n", tmpperf1, tmpperf2, ts);*/
}

static void add_call_event(log_timestamp_t ts, char *call_id, 
		char *caller_uri, char *receiver_uri, int call_event_type,
		int code) {
	struct call *cp, *icp, *prev_cp;

	/* first, try to find a matching existing call */
	cp = calls;
	prev_cp = NULL;
	icp = NULL;
	while (cp != NULL) {
		if (!strcmp(cp->call_id, call_id) && 
				!strcmp(cp->caller_uri, caller_uri)
				&&
				(call_event_type == INVITE_RECEIVED || 
				 cp->call_events == (1 << INVITE_RECEIVED) ||
				 !strcmp(cp->receiver_uri, receiver_uri))) {
			icp = cp;
			break;
		}
		prev_cp = cp;
		cp = cp->next_call;
	}
	if (!icp) {
		if ((cp = (struct call *)malloc(sizeof(struct call))) == NULL) {
			fprintf(stderr, "add_call: Out of memory.\n");
			exit(1);
		}
		strcpy(cp->call_id, call_id);
		strcpy(cp->caller_uri, caller_uri);
		strcpy(cp->receiver_uri, receiver_uri);
		cp->call_events |= (1 << call_event_type);
		cp->event[call_event_type].timestamp = ts;
		cp->event[call_event_type].count = 1;
		cp->event[call_event_type].code = code;
		cp->next_call = calls;
		calls = cp;
		num_calls++;
	}
	else {
		/* if only invite received, receiver field isn't correct yet*/
		if (icp->call_events == (1 << INVITE_RECEIVED)) {
			strcpy(icp->receiver_uri, receiver_uri);
		}
		if (icp->call_events & (1 << call_event_type)) {
			icp->event[call_event_type].count++;
			if (code != icp->event[call_event_type].code) {
				fprintf(stderr, "warning: different code (was: %d new: %d) (call_id: %s event type: %d).\n", icp->event[call_event_type].code, code, call_id, call_event_type);
			}
			/* if only INVITE received and this is a repeat,
			 * update info to save the "closest" invite */
			if (icp->call_events == (1 << INVITE_RECEIVED) &&
				call_event_type == INVITE_RECEIVED) {
				icp->event[call_event_type].code = code;
				icp->event[call_event_type].timestamp = ts;
			}
		}
		else {
			icp->call_events |= (1 << call_event_type);
			icp->event[call_event_type].count = 1;
			icp->event[call_event_type].code = code;
			icp->event[call_event_type].timestamp = ts;
		}
		/* note: retries will produce false incomplete calls, which
		 * are post-processed out in postprocess_calls() */
		if (icp->call_events == ALL_EVENTS_MASK) {
			if (prev_cp) {
				prev_cp->next_call = icp->next_call;
				num_calls--;
			}
			else if (icp == calls) {
				calls = icp->next_call;
				num_calls--;
			} 
			else 
				fprintf(stderr, "\nwarning: add_call oddity.\n");
			icp->next_call = completed_calls;
			completed_calls = icp;
			num_completed_calls++;
		}
	}	
}

/*
 * parser stuff
 */
static void format_process(char *msg, char *proc_entry) {
	char *p = msg;
	char *p_start, *p_end;
	char *m_start;
	while (*p != '\0' && *p != ':') p++;
	if (p == msg || *p == '\0')
		return;
	m_start = p;
	m_start++;
	while (*m_start != '\0' && (*m_start == ' ' || *m_start == '\t'))
		m_start++;
	p--;
	if (*p == ']') {
		while (*p != '[' && p != msg) p--;
		if (p == msg)
			return;
		p--;  
	}
	p_end = p;
	while (*p != '/' && p != msg) p--;
	if (*p == '/') p++;
	p_start = p;
	for (p = p_start; p <= p_end; p++)
		*proc_entry++ = *p;
	*proc_entry = '\0';
	while (*m_start != '\0')
		*msg++ = *m_start++;
	*msg = '\0';
}

static int parse_lookup(log_timestamp_t ts, const char *msg) {
	static char field1[128]; // non-thread safe
	static char field2[128];
	static char field3[128];
	static char field4[128];
	static char field5[128];
	static char field6[128];
	int err;
	char *p, *p_start, *p_end;
	log_timeinterval_t time;

	err = sscanf(msg, "%s %s %s %s %s %s", field1, field2, field3, 
		field4, field5, field6);
	if (err != 6)  { 
		/*printf("parse_lookup: scan() failed: err=%d field1=[%s]\n", err, field1);*/
		return -1;
	}
	if (strcmp(field1, "REGISTRAR:")) {
		/*printf("parse_lookup: not a lookup: field1=[%s]\n", field1);*/
		return -1;
	}
	if (strcmp(field2, "lookup()")) {
		fprintf(stderr, "warning: parse_lookup: field2=[%s]\n", field2);
		return -1;
	}
	/* time */
	if (field6[0] != '[') {
		fprintf(stderr, "Warning: malformed REGISTRAR log entry (time1).\n");			return -1;
	}
	p_start = (char *)&field6[1];
	p = field6;
	while (*p != '\0') p++;
	while (*p != ']' && p != field6) p--;
	if (p == field6 || *p != ']') {
		fprintf(stderr, "Warning: malformed REGISTRAR log entry (time2).\n");
		return -1;
	}
	p_end = p;
	*p_end = '\0';
	time = strtoull(p_start, NULL, 10);	
	
	/* username */
	p_start = (char *)&field4[1];
	p = field4;
	while (*p != '\0') p++;
	while (*p != '\'' && p != field4) p--;
	if (p == field4 || *p != '\'') {
		fprintf(stderr, "Warning: malformed REGISTRAR log entry (username).\n");
		return -1;
	}
	p_end = p;
	*p_end = '\0';

	add_lookup(ts, p_start, time);
	return 0;
}

static int parse_acc(log_timestamp_t ts, const char *msg) {
	static char field1[128]; // non-thread safe
	static char field2[128];
	static char field3[128];
	static char field4[128];
	static char field5[128];
	static char field6[128];
	static char field7[128];
	static char field8[128];
	static char field9[128];
	static char field10[128];
	int err;
	char *p, *p_start, *p_end;
	int call_event_type;
	char *caller_uri, *receiver_uri, *call_id;
	int call_event_code;

	err = sscanf(msg, "%s %s %s %s %s %s %s %s %s %s", field1, field2,
		field3, field4, field5, field6, 
		field7, field8, field9, field10);
	if (err != 10)  { 
		fprintf(stderr, "parse_acc: scan() failed: err=%d field1=[%s]\n", err, field1);
		return -1;
	}
	if (strcmp(field1, "ACC:")) {
		fprintf(stderr, "parse_acc: not ACC entry: field1=[%s]\n", field1);
		return -1;
	}

	/* request accounted */
	if (!strcmp(field2, "request") && !strcmp(field3, "accounted:") &&
			!strcmp(field4, "method=INVITE,")) {
		call_event_type = INVITE_RECEIVED;	
	}
	else if (!strcmp(field2, "transaction") && !strcmp(field3, "answered:")
			&& !strcmp(field4, "method=INVITE,")) {
		call_event_type = INVITE_REPLIED;
	}
	else if (!strcmp(field2, "request") && !strcmp(field3, "acknowledged:")
			&& !strcmp(field4, "method=ACK,")) {
		call_event_type = ACK_RECEIVED;
	}
	else if (!strcmp(field2, "transaction") && !strcmp(field3, "answered:")
			&& !strcmp(field4, "method=BYE,")) {
		call_event_type = BYE_REPLIED;
	}
	else {
		fprintf(stderr, "parse_acc: unknown message.\n");
		return -1;
	}	

	/* strip assumed trailing ',' */

	/* get i-uri, o-uri, call_id and code*/
	/* (can reuse field1,2,3,4) */
	if (!strncmp(field5, "i-uri=", 6)) {
		p_start = (char *)&field5[6];
		p = p_start;
		while (*p != '\0') p++;
		while (*p != ',' && p != p_start) p--;
		if (p == p_start || *p != ',') {
			fprintf(stderr, "Warning: malformed ACC log entry (i-uri).\n");
			return -1;
		}
		p_end = p;
		*p_end = '\0';
		caller_uri = p_start;
	} 
	else {
		fprintf(stderr, "parse_acc: Unexpected field (should be i-uri)\n");
		return -1;
	}
	
	if (!strncmp(field6, "o-uri=", 6)) {
		p_start = (char *)&field6[6];
		p = p_start;
		while (*p != '\0') p++;
		while (*p != ',' && p != p_start) p--;
		if (p == p_start || *p != ',') {
			fprintf(stderr, "Warning: malformed ACC log entry (o-uri).\n");
			return -1;
		}
		p_end = p;
		*p_end = '\0';
		receiver_uri = p_start;
	} 
	else {
		fprintf(stderr, "parse_acc: Unexpected field (should be o-uri)\n");
		return -1;
	}

	if (!strncmp(field7, "call_id=", 8)) {
		p_start = (char *)&field7[8];
		p = p_start;
		while (*p != '\0') p++;
		while (*p != ',' && p != p_start) p--;
		if (p == p_start || *p != ',') {
			fprintf(stderr, "Warning: malformed ACC log entry (call_id).\n");
			return -1;
		}
		p_end = p;
		*p_end = '\0';
		call_id = p_start;
	} 
	else {
		fprintf(stderr, "parse_acc: Unexpected field (should be call_id)\n");
		return -1;
	}

	if (!strncmp(field10, "code=", 5)) {
		p_start = (char *)&field10[5];
		p = p_start;
		while (*p != ' ' && *p != '\n' && *p != '\t' && *p != '\0') p++;
		p_end = p;
		*p_end = '\0';
		if (!strcmp(p_start, "Muhaha"))
			call_event_code = -1;
		else
			call_event_code = atoi(p_start);
	} 
	else {
		fprintf(stderr, "parse_acc: Unexpected field (should be o-uri)\n");
		return -1;
	}

/*	printf("parse_acc: success! [%d] [%s] [%s] [%s] [%d]\n", 
		call_event_type, caller_uri,	
		receiver_uri, call_id, call_event_code);
*/
	add_call_event(ts, call_id, caller_uri, receiver_uri, call_event_type,
			call_event_code);

	return 0;
}

static void parse_message(const char *msg) {
	const char *p = msg;
	const char *p_start, *p_end;
	char tmp[512];
	log_timestamp_t ts;
	int i;

	/* timestamp */
	if (*p != '[') return;
	p_start = ++p;
	if (*p == ']') return;
	
	while (*p != ']' && *p != '\0') p++;
	if (*p != ']') return;
	p_end = p-1;
	for (i = 0; i < p_end-p_start+1; i++)
		tmp[i] = p_start[i];
	tmp[i] = '\0';
	ts = strtoull(tmp, NULL, 10);
	if (ts == 0 && errno) {
		fprintf(stderr, "error in strtoull, errno=%d\n", errno);
		exit(1);
	}
	p++;
	/*printf("parsed [%s] [%lld]\n", tmp, ts);*/
	if (!parse_lookup(ts, p))
		return;
	else if (!parse_acc(ts, p))
		return;
	else {
		fprintf(stderr, "Unrecognised message.\n");
		return;
	}
}


int getline(char s[], int lim);
void debug_print_lookups();
void debug_print_calls(struct call *first_entry, int pi);
void print_help();

int main(int argc, char *argv[]) {
	struct log_event ev;
	char cur_line[LINE_BUFFER_SIZE];
	char prev_line[LINE_BUFFER_SIZE];
	int repeat;
	struct tm logdate;
	char logdatestr[LINE_BUFFER_SIZE];
	time_t begin_time=0, end_time=0;

	int c;
	extern char *optarg;
	extern int optind, optopt;

	cmd_option_output_basename[0] = '\0';
	while ((c = getopt(argc, argv, "cdhqlo:u:b:e:")) != -1) {
		switch(c) {
		case 'q':
			cmd_option_quiet++;
			break;
		case 'l':
			cmd_option_list_calls = 1;
			break;
		case 'h':
			print_help();
			exit(1);
			break;
		case 'u':
			cmd_option_calls_per_unit = strtoull(optarg, NULL, 10);
			break;
		case 'c':
			cmd_option_output_csv = 1;
			break;
		case 'd':
			cmd_option_drop_tsfaults = 1;
			break;
		case 'o':
			if (strlen(optarg) == 0) {
				printf("Illegal basename.\n");
				print_help();
				exit(1);
			}
			strcpy(cmd_option_output_basename, optarg);	
			break;
		case 'b':
			{
			struct tm *tmp_tm;
			tmp_tm = malloc(sizeof(struct tm));
			strptime(optarg, CMDDATEFMT, tmp_tm);
			if (cmd_option_begin_date)
				free(cmd_option_begin_date);
			cmd_option_begin_date = tmp_tm;
                        cmd_option_begin_date->tm_isdst=-1; /* discard automatic daylight saving time option */
			}
			break;
		case 'e':
			{
			struct tm *tmp_tm;
			tmp_tm = malloc(sizeof(struct tm));
			strptime(optarg, CMDDATEFMT, tmp_tm);
			if (cmd_option_end_date)
				free(cmd_option_end_date);
			cmd_option_end_date = tmp_tm;
                        cmd_option_end_date->tm_isdst=-1; /* discard automatic daylight saving time option */
			}
			break;
		case '?':
		default:
			printf("Error in command-line options (-%c)\n", optopt);
			print_help();
			exit(1);
			break;
		}
	}

	if (cmd_option_begin_date) {
		begin_time = mktime(cmd_option_begin_date);
		if (begin_time == -1) {
			printf("Error parsing begin date.\n");
			print_help();
			exit(1);
		}
	}
	if (cmd_option_end_date) {
		end_time = mktime(cmd_option_end_date);
		if (end_time == -1) {
			printf("Error parsing end date.\n");
			print_help();
			exit(1);
		}
	}

	if (strlen(cmd_option_output_basename)) {
		char tmpname[512];
		FILE *f;
		strcpy(tmpname, cmd_option_output_basename);
		if (cmd_option_output_csv)
			strcat(tmpname, "-calls.csv");
		else
			strcat(tmpname, "-calls.octave");
		if ((f = fopen(tmpname, "w")) == NULL) {
			printf("Error opening output: %s\n", tmpname);
			exit(1);
		}
		cmd_option_call_output = f;
		strcpy(tmpname, cmd_option_output_basename);
		if (cmd_option_output_csv)
			strcat(tmpname, "-cpu.csv");
		else
			strcat(tmpname, "-cpu.octave");
		if ((f = fopen(tmpname, "w")) == NULL) {
			printf("Error opening output: %s\n", tmpname);
			fclose(cmd_option_call_output);
			exit(1);
		}
		cmd_option_cps_output = f;	
		strcpy(tmpname, cmd_option_output_basename);
		if (cmd_option_output_csv)
			strcat(tmpname, "-lookups.csv");
		else
			strcat(tmpname, "-lookups.octave");
		if ((f = fopen(tmpname, "w")) == NULL) {
			printf("Error opening output: %s\n", tmpname);
			fclose(cmd_option_cps_output);
			fclose(cmd_option_call_output);
			exit(1);
		}
		cmd_option_lookup_output = f;	
	}

	lookups = NULL; prev_added_lookup_pred = NULL;
	calls = NULL; completed_calls = NULL;
	prev_line[0] = '\n';
	prev_line[1] = '\0';
	repeat = 0;
	num_lines = 0;
	while(1) {
		int err;
		time_t etime=0;

		if (!repeat) {
			if (!getline(cur_line, LINE_BUFFER_SIZE)) {
				break;
			}
			num_lines++;
		}
		else  {
			repeat--;
		}
		if (!cmd_option_quiet && (num_lines % 500) == 0) {
			printf("lines read [%llu] lookups: %llu calls: %llu complete: %llu\r", num_lines, num_lookups, num_calls, num_completed_calls);
			fflush(stdout);
		}
		err = sscanf(cur_line, "%s %s %s %s %[^\n]", ev.month, ev.day, ev.time, 
			ev.host, ev.message);
		sprintf(logdatestr, "%s %s %s", ev.month, ev.day, ev.time);
		if (cmd_option_begin_date)
			logdate.tm_year = cmd_option_begin_date->tm_year;
		else if (cmd_option_end_date)
			logdate.tm_year = cmd_option_end_date->tm_year;
		if (cmd_option_begin_date || cmd_option_end_date) {
			strptime(logdatestr, LOGDATEFMT, &logdate);
			etime = mktime(&logdate);
			if (etime == -1) {
				fprintf(stderr, "Error parsing log date at line %s\n", cur_line);
				exit(1);
			}
		}
		if (cmd_option_begin_date && 
			difftime(etime, begin_time) < 0)
			continue;
		if (cmd_option_end_date && 
			difftime(end_time, etime) < 0)
			continue;

		if (err != 5) {	
			fprintf(stderr, "warning: skipping broken log entry.\n");
			fprintf(stderr, "[%s]\n", cur_line);
			continue;
		}
		/* is it "last message repeated N times" */
		if (!strncmp(ev.message, "last message repeated", 21)) {
			repeat = atoi(ev.message+22);
			strcpy(cur_line, prev_line);
			continue;
		}
		strcpy(prev_line, cur_line);
		ev.process[0] = '\0';
		format_process(ev.message, ev.process);
		if (strcmp(ev.process, OPENSER_PNAME)) {
			continue;
		}
		parse_message((char *)&ev.message);
	}

	if (cmd_option_quiet < 2) {
		printf("lines read [%llu] lookups: %llu calls: %llu complete: %llu\n", num_lines, num_lookups, num_calls, num_completed_calls);
		fflush(stdout);
	}

#ifdef DBG
	debug_print_calls(calls, 1);
#endif
	if (cmd_option_quiet < 2)
		printf("\npostprocessing...\n"); fflush(stdout);
	postprocess_calls();
#ifdef DBG
	debug_print_lookups();
	debug_print_calls(completed_calls);
	debug_print_calls(calls, 1);
#endif
	if (cmd_option_quiet < 2)
		printf("\ngenerating stats...\n"); fflush(stdout);
	generate_lookup_stats();
	generate_call_stats();

	if (cmd_option_lookup_output)
		fclose(cmd_option_lookup_output);
	if (cmd_option_cps_output)
		fclose(cmd_option_cps_output);
	if (cmd_option_call_output)
		fclose(cmd_option_call_output);
	
	return 0;
}

void debug_print_lookups() {
	long long lookup_count = 0;
	struct lookup *lp;
	struct lookup *prev_lp;
	printf("debug_print_lookups():");
	prev_lp = NULL;
	lp = lookups;
	printf("\n  lookups @%p:\n\n", lookups);
	while (lp != NULL) {
		if (prev_lp != NULL && prev_lp->timestamp > lp->timestamp)
			printf("warning: timestamp ordering error at %lld: %llu > %llu\n", lookup_count, prev_lp->timestamp, lp->timestamp);
		printf("    (%lld) %llu %s %llu\n", lookup_count, lp->timestamp, lp->username, lp->time);
		if (lp->time < 0 || lp->time > 5000)
			printf("warning: odd lookup time %llu\n", lp->time);
		lookup_count++;	
		prev_lp = lp;
		lp = lp->next_lookup;
	}
}

void debug_print_calls(struct call *first_entry, int pi) {
	long long call_count = 0;
	struct call *cp;
	struct call *prev_cp;
	printf("debug_print_calls():");
	cp = first_entry;
	printf("\n  calls @%p:\n\n", cp);
	while (cp != NULL) {
		printf("    (%lld)", pi ? call_count : 0LL);
		print_call(stdout, cp);
		call_count++;	
		prev_cp = cp;
		cp = cp->next_call;
	}
}

void print_help() {
	printf("usage:\n");
	printf("alog [OPTIONS] < myfile\n");
	printf("-q            suppress progress output (can be repeated)\n");
	printf("-l            list details for failed calls\n");
	printf("-o<basename>  output lookups and calls to\n");
	printf("              <basename>-calls.csv - all calls\n");
	printf("              <basename>-cpu.csv - calls-per-unit\n");
	printf("              <basename>-lookups.csv - all lookups\n");
	printf("-u<time>      unit for calls-per-unit in ms (default: %llu ms)\n",
		DEFAULTCALLSPERUNIT);
	printf("-c            use CSV output format (default is octave)\n");
	printf("-d            drop calls with time stamp order errors from output\n");
	printf("-b<date>      begin date for log parsing (inclusive)\n");
	printf("-e<date>      end date for log parsing (inclusive)\n");
	printf("date format: \"YYYY MMM DD hh:mm:ss\"\n");
	printf("example: alog -o onesec -b \"2006 Nov 1 15:47:25\" -e \"2006 Nov 1 15:47:25\"\n");
}

int getline(char s[], int lim) {
	int c=0,i;
	for (i=0; i<lim-1 && (c=getchar()) != EOF && c!='\n'; ++i)
		s[i] = c;
	if (c == '\n') {
		s[i] = c;
		++i;
	}
	s[i] = '\0';
	return i;
}

