/* To compile:
   gcc -I${OSIPDIR}/include -L${OSIPDIR}/lib sip_reg.c \
       -o sip_reg -leXosip2 -losip2 -osipparser2 -lpthread

   Add -DENABLE_DEBUG if you want.
*/

/* Based on the code from libeXosip2-2.2.3/tools/sip_reg.c */

#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <eXosip2/eXosip.h>

#define _GNU_SOURCE
#ifndef _GNU_SOURCE
#error "Not GNU source"
#endif /* _GNU_SOURCE hassle */
#include <getopt.h>

#ifdef ENABLE_DEBUG
static void *reg_listener_debug(void *arg);
#endif /* ENABLE_DEBUG */
static void *reg_listener(void *arg);

typedef struct regparam_t {
	int regid;
	int expiry;
	int auth;
} regparam_t;

#ifdef ENABLE_DEBUG
static void *reg_listener_debug(void *arg)
{
	eXosip_event_t *event;
	char *msg, *start;
	int array[1000], index, rem;
	unsigned int len;

	if (eXosip_listen_addr(IPPROTO_UDP, NULL, *(int *)arg, AF_INET, 0)) {
		perror("eXosip_listen_addr");
		pthread_exit(NULL);
	}

	for (index = 0; index < 1000; index++) array[index] = 0;

	for (;;) {
		event = eXosip_event_wait(0, 0);
		if (!event) {
			usleep(100000);
			continue;
		}
		osip_message_to_str(event->response, &msg, &len);
		if (event->type == EXOSIP_REGISTRATION_NEW ||
		    event->type == EXOSIP_REGISTRATION_SUCCESS) {
			/* fprintf(stderr, "GOT: %s\n", msg); */
			start = strstr(msg, "\nFrom:");
			if (!start) {
				fprintf(stderr, "BAD RESPONSE\n");
				continue;
			}
			start = strchr(start, '.');
			start++;
			index = atoi(start);
			fprintf(stderr, "DBG: %d == [%.20s].\n", index, start);
			array[index] = 1;
			for (rem = 1000, index = 0; index < 1000; index++)
				if (array[index]) rem--;
			fprintf(stderr, "DBG: remaining=%d\n", rem);
			if (!rem) break;
			continue;
		}
		fprintf(stderr, "UNKNOWN eXosip response %s\n", msg);
	}
	pthread_exit(NULL);
}
#endif /* ENABLE_DEBUG */

static void *reg_listener(void *arg)
{
	eXosip_event_t *event;
	int success = 0;

	if (eXosip_listen_addr(IPPROTO_UDP, NULL, *(int *)arg, AF_INET, 0)) {
		perror("eXosip_listen_addr");
		pthread_exit(NULL);
	}

	for (;;) /* ever */ {
		event = eXosip_event_wait(0, 0);
		if (!event) {
			usleep(100000);
			continue;
		}
		if (event->type == EXOSIP_REGISTRATION_NEW ||
		    event->type == EXOSIP_REGISTRATION_SUCCESS) {
			success++;
			if (success % 1000 == 0)
				printf("Done %d REGISTER messages.\n", success);
			continue;
		}
		if (event->type == EXOSIP_REGISTRATION_FAILURE)
			/* XXX We assume only bogus retransmits fail. */
			continue;
		fprintf(stderr,
		        "UNKNOWN eXosip response (%s)\n", event->textinfo);
	}
	pthread_exit(NULL);
}

int main(int argc, char **argv)
{
	static struct option lo[] = {
		{ "contact", required_argument, NULL, 'c' },
		{ "expiry", required_argument, NULL, 'e' },
		{ "from", required_argument, NULL, 'f' },
		{ "maxid", required_argument, NULL, 'n' },
		{ "proxy", required_argument, NULL, 'p' },
		{ "port", required_argument, NULL, 'P' },
	};
	struct regparam_t regparam = { 0, 3600, 0 };
	struct servent *service;
	struct timespec onemsec;
	static char *so = "c:e:f:n:p:P:";
	char *contact = NULL, *from = NULL, *proxy = NULL;
	char *contactuser = NULL, *fromuser = NULL;
	pthread_t listener_thread;
	osip_message_t *reg = NULL;
	int expiry = 3600, maxid = 0, port = 5060;
	int c, i, option_index, res;

	for (;;) /* all options */ {
		c = getopt_long(argc, argv, so, lo, &option_index);
		if (c == -1) break;
		switch(c) {
			case 'c':
				contact = optarg;
				break;
			case 'e':
				expiry = atoi(optarg);
				break;
			case 'f':
				from = optarg;
				break;
			case 'n':
				maxid = atoi(optarg);
				break;
			case 'p':
				proxy = optarg;
				break;
			case 'P':
				service = getservbyname(optarg, "udp");
				if (service) port = service->s_port;
				else port = atoi(optarg);
			default:
				break;
		}
	}

	if (eXosip_init()) {
		perror("eXosip_init");
		return 1;
	}

	if (!contact || !from || !proxy || !maxid) {
		fprintf(stderr, "Incorrect arguments.\n");
		return 1;
	}

#ifdef ENABLE_DEBUG
	if (pthread_create(&listener_thread, NULL, reg_listener_debug, &port))
#else
	if (pthread_create(&listener_thread, NULL, reg_listener, &port))
#endif /* ENABLE_DEBUG */
	{
		perror("pthread_create");
		return 1;
	}

	onemsec.tv_sec = 0;
	onemsec.tv_nsec = 50000000L;

	fromuser = (char *)malloc(strlen(from) + 30); /* XXX should use -lm */
	contactuser = (char *)malloc(strlen(contact) + 30);  /* ditto */

	/* FIXME! For whatever reason, we never get reply to first request. */
	for (i = -1; i < maxid; i++) {
		snprintf(fromuser, strlen(from) + 29, from, i);
		snprintf(contactuser, strlen(contact) + 29, contact, i);
#ifdef ENABLE_DEBUG
		fprintf(stderr, "DBG: fromuser=%s, contactuser=%s\n", fromuser, contactuser);
#endif /* ENABLE_DEBUG */
		eXosip_lock();
		regparam.regid = eXosip_register_build_initial_register(fromuser, proxy, contactuser, expiry, &reg);
		res = eXosip_register_send_register(regparam.regid, reg);
		if (res != 0 && errno != 0) {
			perror("eXosip_register_send_register");
			eXosip_unlock();
			return 2;
		}
		eXosip_unlock();
		if (i % 100 == 0)
			printf("Sent %d REGISTER messages.\n", i);
		nanosleep(&onemsec, NULL);		
	}
	printf("Sent %d REGISTER messages.\n", maxid);
	return pthread_join(listener_thread, NULL);
}
